#pragma rtGlobals=1		// Use modern global access method.


strconstant K_FFTWindowList="Bartlett;Blackman;cos2;Hamming;Hanning;KaiserBessel;Poisson2;Riemann"
constant K_specmaxparams=3
//References:

//Gardner, W.A., Statistical Spectral Analysis: A Nonprobabilistic Theory, 
//Prentice Hall, Englewood Cliffs, NJ, 1988, 566 pp.

//Bloomfield, P., Fourier Analysis of Time Series:  An Introduction,
//John Wiley & Sons, New York, 1976, 258 pp.

Structure SpectrumParameters
	//Universal structure to be used for input parameters to any function producing a spectral estimate
	
	wave dw,xw				//input waves for XY-pair data
	wave dw_int				//input wave for waveform data
	wave magnitude,phase		//output waves
	variable getParams		//set to 1 when calling to retrieve info for creation of controls
	
	variable Xmin,Xmax		//minimum and maximum x values to define a subrange of the data
	variable Fmin,Fmax		//minimum and maximum frequency of interest
	variable freqRes			//desired frequency resolution; if 0, default resolution (1/series length)
								//if <1/length, data will be padded to obtain desired resolution
	
	struct SpecSetVar SV			//structures holding set-variable parameters specific to a spectral estimate method
	
	struct SpecPopMenu PM		//structures holding popup menu parameters specific to a spectral estimate method
endStructure

Structure SpecSetVar
	//parameters to create a set variable control
	string name[K_specmaxparams]				//control name
	variable value	[K_specmaxparams]			//value
	variable low[K_specmaxparams],high[K_specmaxparams]	//low and high limits and increment for creating set variable control
	variable inc[K_specmaxparams]	
endStructure

Structure SpecPopMenu
	//parameters to create a popup menu control
	string name[K_specmaxparams]				//control name
	string value[K_specmaxparams]				//semicolon-separated list for popup menu
	string popStr[K_specmaxparams]				//value of selected item
	variable popNum[K_specmaxparams]			//number of selected item
endStructure

Function SPECTRUM_prototype(SPs)
	struct SpectrumParameters &SPs
	
	if(SPs.getParams)
		SPS.SV.name[0]="";SPS.SV.value[0]=nan;SPS.SV.low[0]=nan;SPS.SV.high[0]=nan;SPS.SV.inc[0]=nan
		SPS.SV.name[1]="";SPS.SV.value[1]=nan;SPS.SV.low[1]=nan;SPS.SV.high[1]=nan;SPS.SV.inc[1]=nan
		SPS.SV.name[2]="";SPS.SV.value[2]=nan;SPS.SV.low[2]=nan;SPS.SV.high[2]=nan;SPS.SV.inc[2]=nan
		SPS.PM.name[0]="";SPS.PM.value[0]="";SPS.PM.popStr[0]="";SPS.PM.popNum[0]=nan
		SPS.PM.name[1]="";SPS.PM.value[1]="";SPS.PM.popStr[1]="";SPS.PM.popNum[1]=nan
		SPS.PM.name[2]="";SPS.PM.value[2]="";SPS.PM.popStr[2]="";SPS.PM.popNum[2]=nan
	else
//		SPs.magnitude=nan;SPs.phase=nan
	endif	
end

Function SPECTRUM_Periodogram(SPs)
	struct SpectrumParameters &SPs

	if(SPs.getParams)
		SPS.SV.name[0]="";SPS.SV.value[0]=nan;SPS.SV.low[0]=nan;SPS.SV.high[0]=nan;SPS.SV.inc[0]=nan
		SPS.SV.name[1]="";SPS.SV.value[1]=nan;SPS.SV.low[1]=nan;SPS.SV.high[1]=nan;SPS.SV.inc[1]=nan
		SPS.SV.name[2]="";SPS.SV.value[2]=nan;SPS.SV.low[2]=nan;SPS.SV.high[2]=nan;SPS.SV.inc[2]=nan
		SPS.PM.name[0]="Taper Function"
		SPS.PM.value[0]="none;"+K_FFTWindowList
		SPS.PM.popStr[0]="Hamming"
		SPS.PM.popNum[0]=whichlistitem(SPS.PM.popStr[0],SPS.PM.value[0])+1
		SPS.PM.name[1]="";SPS.PM.value[1]="";SPS.PM.popStr[1]="";SPS.PM.popNum[1]=nan
		SPS.PM.name[2]="";SPS.PM.value[2]="";SPS.PM.popStr[2]="";SPS.PM.popNum[2]=nan
	else
		pgram(SPs.dw_int,pgramMagS=SPs.magnitude,win=SPs.PM.popStr[0],pad=1/SPs.freqRes/deltax(SPS.dw_int))
		SPs.magnitude=sqrt(SPs.magnitude)
	endif
end

Function SPECTRUM_Welch(SPs)
	struct SpectrumParameters &SPs

	if(SPs.getParams)
		SPS.SV.name[0]="number of segments";SPS.SV.value[0]=5;SPS.SV.low[0]=1;SPS.SV.high[0]=inf;SPS.SV.inc[0]=1
		SPS.SV.name[1]="percent overlap";SPS.SV.value[1]=50;SPS.SV.low[1]=0;SPS.SV.high[1]=100;SPS.SV.inc[1]=5
		SPS.SV.name[2]="";SPS.SV.value[2]=nan;SPS.SV.low[2]=nan;SPS.SV.high[2]=nan;SPS.SV.inc[2]=nan
		SPS.PM.name[0]="Taper Function"
		SPS.PM.value[0]="none;"+K_FFTWindowList
		SPS.PM.popStr[0]="Hamming"
		SPS.PM.popNum[0]=whichlistitem(SPS.PM.popStr[0],SPS.PM.value[0])+1
		SPS.PM.name[1]="";SPS.PM.value[1]="";SPS.PM.popStr[1]="";SPS.PM.popNum[1]=nan
		SPS.PM.name[2]="";SPS.PM.value[2]="";SPS.PM.popStr[2]="";SPS.PM.popNum[2]=nan
	else
		newdatafolder/o root:Packages
		newdatafolder/o root:Packages:WelchSpectrum
		make/o/d root:Packages:WelchSpectrum:tmp_dw,root:Packages:WelchSpectrum:tmp_magS
		wave tmp_dw=root:Packages:WelchSpectrum:tmp_dw,tmp_magS=root:Packages:WelchSpectrum:tmp_magS
		variable i,npnts,nsegs,seglen,segoffset,segolf
		npnts=numpnts(SPs.dw_int)
		nsegs=SPs.SV.value[0]
		segolf=SPs.SV.value[1]/100
		seglen=npnts/(nsegs-nsegs*segolf+segolf)
		segoffset=seglen*(1-segolf)
		for(i=0;i<nsegs;i+=1)
			duplicate/r=[i*segoffset,i*segoffset+seglen-1]/o SPs.dw_int,tmp_dw
			pgram(tmp_dw,pgramMagS=tmp_magS,win=SPs.PM.popStr[0],pad=1/SPs.freqRes/deltax(tmp_dw))
			if(i==0)
				duplicate/o tmp_magS,SPs.Magnitude
			else
				SPs.magnitude+=tmp_magS
			endif
		endfor
		SPs.magnitude=sqrt(SPs.magnitude/nsegs^2)
		copyscales/p tmp_magS,sps.magnitude
		killdatafolder/z "root:Packages:WelchSpectrum"
	endif
end

Function Pgram(dw[,pgramMagS,win,pad])
	wave dw,pgramMagS
	string win
	variable pad
		
	string PgramName
	variable np,np2,magfac
	if(paramisdefault(pgramMagS))
		PgramName=GetWavesDataFolder(dw,2)+"_pgram"
	else
		PgramName=GetWavesDataFolder(pgramMagS,2)
	endif
	make/o/d $PgramName
	wave pgramMagS=$PgramName
	duplicate/o dw,pgramMagS
	wavestats/M=1/q dw
	np=v_npnts
	pgramMagS-=v_avg
	if(ParamIsDefault(pad) || (pad==0) || numtype(pad) || np>pad)
		np2=np+mod(np,2)
	else
		pad=round(pad)
		np2=pad+mod(pad,2)
//		np2=2^(ceil(log(np2)/log(2))+1)
	endif
	if(ParamIsDefault(win) || whichlistitem(win,K_FFTWindowList)<0)
		fft/PAD={np2}/DEST=$PgramName/MAGS pgramMagS
	else
		fft/PAD={np2}/DEST=$PgramName/MAGS/WINF=$win pgramMagS
		magfac=winNorm(np,win)
		pgramMagS/=magfac
	endif
	pgramMagS/=(np/2)^2
end

function winNorm(n,windowF)
	variable n
	string windowF
	
	newdatafolder/o root:Packages
	newdatafolder/o root:Packages:winNorm
	make/o/n=(n) root:Packages:winNorm:tmpWin
	wave tmpWin=root:Packages:winNorm:tmpWin
	tmpWin=1
	windowFunction/FFT=1 $windowF,tmpWin
	wavestats/q tmpWin
	killdatafolder/Z root:Packages:winNorm
	return v_rms^2
end

Function SPECTRUM_BlackmanTukey(SPs)
	struct SpectrumParameters &SPs
	
	if(SPs.getParams)
		SPS.SV.name[0]="Lag fraction"
		SPS.SV.value[0]=0.5
		SPS.SV.low[0]=0.05
		SPS.SV.high[0]=1
		SPS.SV.inc[0]=0.05
		SPS.SV.name[1]="";SPS.SV.value[1]=nan;SPS.SV.low[1]=nan;SPS.SV.high[1]=nan;SPS.SV.inc[1]=nan
		SPS.SV.name[2]="";SPS.SV.value[2]=nan;SPS.SV.low[2]=nan;SPS.SV.high[2]=nan;SPS.SV.inc[2]=nan
		SPS.PM.name[0]="Taper Function"
		SPS.PM.value[0]="none;"+K_FFTWindowList
		SPS.PM.popStr[0]="Hamming"
		SPS.PM.popNum[0]=whichlistitem(SPS.PM.popStr[0],SPS.PM.value[0])+1
		SPS.PM.name[1]="";SPS.PM.value[1]="";SPS.PM.popStr[1]="";SPS.PM.popNum[1]=nan
		SPS.PM.name[2]="";SPS.PM.value[2]="";SPS.PM.popStr[2]="";SPS.PM.popNum[2]=nan
	else
		BTspec(SPs.dw_int,dwMag2Name=getWavesDataFolder(SPs.magnitude,2),lagfrac=SPS.SV.value[0],winF=SPs.PM.popStr[0],pad=1/SPs.freqRes/deltax(SPs.dw_int))
		SPs.magnitude=sqrt(SPs.magnitude)
	endif	
end

Function BTspec(dw,[dw2,lagfrac,pad,winF,dwMag2Name,dwPhName,dw2Mag2name,dw2PhName,XMag2Name,XPhName,XCohName])
	wave dw,dw2;variable lagfrac,pad
	string dwMag2Name,dwPhName,dw2Mag2name,dw2PhName
	string winF,XMag2Name,XPhName,XCohName
	
	string dwFolder="",dwName="",dw2Name="",dw2Folder=""
	variable dw_avg,dw2_avg,x1,x2,dx,var,var2,cov,np,npad,lag
	
	dwFolder=GetWavesDatafolder(dw,1);dwName=nameofwave(dw)
	if(paramisdefault(dwMag2Name))
		dwMag2Name=dwFolder+dwName+"_BTMag2"
	endif
	if(paramisdefault(dwPhName))
		dwPhName=dwFolder+dwName+"_BTphase"
	endif
	if(ParamIsDefault(dw2))
		wave dw2=$getwavesdatafolder(dw,2)
	else
		dw2name=nameofwave(dw2)
		dw2Folder=getwavesdatafolder(dw2,1)
		if(ParamIsDefault(dw2Mag2Name))
			dw2Mag2Name=dw2Folder+dw2Name+"_BTMag2"
		endif
		if(ParamIsDefault(dw2PhName))
			dw2PhName=dw2Folder+dw2Name+"_BTphase"
		endif
		if(ParamIsDefault(XMag2Name))
			XMag2Name=dwFolder+dwName+"_"+dw2Name+"_XBTMag2"
		endif
		if(ParamIsDefault(XPhName))
			XPhName=dwFolder+dwName+"_"+dw2Name+"_XBTphase"
		endif
		if(ParamIsDefault(XCohName))
			XCohName=dwFolder+dwName+"_"+dw2Name+"_XBTcoh"
		endif
	endif
	NewDataFolder/o/s TMP_BTspec
	dx=max(abs(deltax(dw)),abs(deltax(dw2)))
	x1=max(pnt2x(dw,0),pnt2x(dw2,0))
	x2=min(pnt2x(dw,numpnts(dw)-1),pnt2x(dw2,numpnts(dw2)-1))
	if(x1>x2)
		doalert 0,"Waves don't overlap."
		return 0
	endif
	np=round(abs(x2-x1)/dx+1)
	if(ParamIsDefault(lagfrac))
		lag=round(np*2)
	else
		lag=ceil(lagfrac*(np*2-1))
		lag=lag-mod(lag,2)
	endif
	wavestats/q/r=(x1,x2) dw;dw_avg=v_avg;var=v_sdev^2
	wavestats/q/r=(x1,x2) dw2;dw2_avg=v_avg;var2=v_sdev^2
	cov=covariance(dw,dw2)
	make/d/o/n=(np) dwCorr,dw2Corr,XCorr
	setscale/p x x1,dx,dwCorr,dw2Corr,XCorr
	dwCorr=dw(x)-dw_avg;dw2Corr=dw2(x)-dw2_avg;XCorr=dw2(x)-dw2_avg

//calculate spectrum for dw:
	correlate dwCorr,XCorr;XCorr/=np/4
	correlate dw2Corr,dw2Corr;dw2Corr/=np/4
	correlate dwCorr,dwCorr;dwCorr/=np/4
	
	BTspec_calc(dwCorr,lag,npad=pad,win=winF);wave W_magsqr,W_phase
	duplicate/o W_magsqr,$dwMag2Name
	duplicate/o W_phase,$dwPhName
	wave dwMag2=$dwMag2Name,dwPh=$dwPhName
	//dwPh[0]=dwPh[1]//;unwrap 2*pi,dwPh
	if(strlen(dw2name))
//calculate spectrum for dw2:
		BTspec_calc(dw2Corr,lag,npad=pad,win=winF)
		duplicate/o W_magsqr,$dw2Mag2Name
		duplicate/o W_phase,$dw2PhName
		wave dw2Mag2=$dw2Mag2Name,dw2Ph=$dw2PhName
		//dw2Ph[0]=dw2Ph[1]//;unwrap 2*pi,dw2Ph
//calculate cross spectrum:
		BTspec_calc(XCorr,lag,npad=pad,win=winF)
		duplicate/o W_magsqr,$XMag2Name,$XCohName
		duplicate/o W_phase,$XPhName
		wave XMag2=$XMag2Name,XPh=$XPhName,XCoh=$XCohName
		//XPh[0]=XPh[1]//;unwrap 2*pi,XPh
		XCoh=min((XMag2)/(sqrt(dwMag2)*sqrt(dw2Mag2)),1)
		wavestats/q XMag2;v_max=(sqrt(v_max)*0.1)^2
		XPh=((XMag2>v_max)*(XCoh>0.5)) ? XPh : nan
		XCoh=((XMag2>v_max)*(XCoh>0.5)) ? XCoh : nan
	endif
	
	
	SetDataFolder ::
	killdatafolder TMP_BTspec
end

function BTspec_calc(w,lag[,npad,win])
	wave w;variable lag,npad;string win
	
	variable np,fftsign,magfac
	np=numpnts(w)
	if(ParamIsDefault(npad) || (npad==0) || numtype(npad))
		npad=np+mod(np,2)
	else
		npad=round(2*npad)
		npad=2^(ceil(ln(lag)/ln(2))+1)
	endif
	rotate -floor((np-lag)/2),w;redimension/n=(lag) w
	if(ParamIsDefault(win) || whichlistitem(win,K_FFTWindowList)<0)
		fft/PAD={npad}/DEST=fftw w
		wavetransform magsqr,fftw
		wavetransform phase,fftw
		wave W_magsqr
		W_magsqr=sqrt(W_magsqr)
	else
		fft/PAD={npad}/DEST=fftw/WINF=$win w
		wavetransform magsqr,fftw
		wavetransform phase,fftw
		wave W_magsqr
		magfac=winNorm(lag,win)
		W_magsqr=sqrt(W_magsqr)/magfac
	endif
//	fftw[1,;2]=-1*(fftw)
	
//	fftw=conj(fftw)	//??
	
//	duplicate/o fftw,fft_
	W_magsqr/=2*np
end

Function Covariance(dw1,dw2)
	wave dw1,dw2
	
	variable avg1,avg2,i,np,summ=0
	np=numpnts(dw1)
	if(numpnts(dw2)!=np)
		return nan
	endif
	wavestats/q dw1;avg1=v_avg
	wavestats/q dw2;avg2=v_avg
	for(i=0;i<np;i+=1)
		summ+=(dw1[i]-avg1)*(dw2[i]-avg2)
	endfor
	return summ/(np-1)
end

